@@ -9,7 +9,7 @@ import warnings
 import pandas as pd
 import plotly.io as pio
 
-from typing import Union
+from typing import Union, Type
 from tqdm.auto import tqdm
 from hashlib import md5
 from pathlib import Path
@@ -23,6 +23,7 @@ from collections import defaultdict
 from ratelimit import limits, sleep_and_retry
 from pyisemail.diagnosis import BaseDiagnosis
 from swagger_spec_validator.common import SwaggerValidationError
+from jsonschema.exceptions import ValidationError
 from bravado_core.formatter import SwaggerFormat
 from bravado.client import SwaggerClient
 from bravado.fido_client import FidoClient  # async
@@ -30,6 +31,8 @@ from bravado.http_future import HttpFuture
 from bravado.swagger_model import Loader
 from bravado.config import bravado_config_from_config_dict
 from bravado_core.spec import Spec
+from bravado.exception import HTTPNotFound
+from bravado_core.validate import validate_object
 from json2html import Json2Html
 from IPython.display import display, HTML, Image, FileLink
 from boltons.iterutils import remap
@@ -50,7 +53,7 @@ MAX_BYTES = 200 * 1024
 DEFAULT_HOST = "contribs-api.materialsproject.org"
 BULMA = "is-narrow is-fullwidth has-background-light"
 PROVIDERS = {"github", "google", "facebook", "microsoft", "amazon"}
-COMPONENTS = {"structures", "tables", "attachments"}
+COMPONENTS = ["structures", "tables", "attachments"]  # using list to maintain order
 VALID_URLS = {f"http://{h}:{p}" for h in ["localhost", "contribs-api"] for p in [5000, 5002, 5003]}
 VALID_URLS |= {f"https://{n}-api.materialsproject.org" for n in ["contribs", "lightsources", "ml"]}
 VALID_URLS |= {
@@ -152,26 +155,37 @@ def visit(path, key, value):
     return True
 
 
-def in_ipython():
+def _in_ipython():
     ipython = sys.modules['IPython'].get_ipython()
     return ipython is not None and 'IPKernelApp' in ipython.config
 
 
 class Dict(dict):
-    """custom dictionary to display itself as Bulma table"""
-    def display(self, attrs=f'class="table {BULMA}"'):
+    """Custom dictionary to display itself as HTML table with Bulma CSS"""
+    def display(self, attrs: str = f'class="table {BULMA}"'):
+        """Nice table display of dictionary
+
+        Args:
+            attrs (str): table attributes to forward to Json2Html.convert
+        """
         html = j2h.convert(json=remap(self, visit=visit), table_attributes=attrs)
-        if in_ipython():
+        if _in_ipython():
             return display(HTML(html))
 
         return html
 
 
 class Table(pd.DataFrame):
+    """Wrapper class around pandas.DataFrame to provide display() and info()"""
     def display(self):
-        return self.plot(**self.attrs)
+        """Display a plotly graph for the table if in IPython/Jupyter"""
+        if _in_ipython():
+            return self.plot(**self.attrs)
 
-    def info(self):
+        return self
+
+    def info(self) -> Type[Dict]:
+        """Show summary info for table"""
         info = Dict((k, v) for k, v in self.attrs.items())
         info["columns"] = ", ".join(self.columns)
         info["nrows"] = len(self.index)
@@ -179,10 +193,12 @@ class Table(pd.DataFrame):
 
 
 class Structure(PmgStructure):
+    """Wrapper class around pymatgen.Structure to provide display() and info()"""
     def display(self):
         return self  # TODO use static image from crystal toolkit?
 
-    def info(self):
+    def info(self) -> Type[Dict]:
+        """Show summary info for structure"""
         info = Dict((k, v) for k, v in self.attrs.items())
         info["formula"] = self.composition.formula
         info["reduced_formula"] = self.composition.reduced_formula
@@ -191,11 +207,17 @@ class Structure(PmgStructure):
 
 
 class Attachment(dict):
-    """Class to handle attachments"""
+    """Wrapper class around dict to handle attachments"""
     def decode(self) -> str:
+        """Decode base64-encoded content of attachment"""
         return b64decode(self["content"], validate=True)
 
     def write(self, outdir: Union[str, Path] = None) -> Path:
+        """Write attachment to file using its name
+
+        Args:
+            outdir (str,Path): existing directory to which to write file
+        """
         outdir = outdir or "."
         path = Path(outdir) / self.name
         content = self.decode()
@@ -203,7 +225,12 @@ class Attachment(dict):
         return path
 
     def display(self, outdir: Union[str, Path] = None):
-        if in_ipython():
+        """Display Image/FileLink for attachment if in IPython/Jupyter
+
+        Args:
+            outdir (str,Path): existing directory to which to write file
+        """
+        if _in_ipython():
             if self["mime"].startswith("image/"):
                 content = self.decode()
                 return Image(content)
@@ -214,6 +241,7 @@ class Attachment(dict):
         return self.info().display()
 
     def info(self) -> Dict:
+        """Show summary info for attachment"""
         fields = ["id", "name", "mime", "md5"]
         info = Dict((k, v) for k, v in self.items() if k in fields)
         info["size"] = len(self.decode())
@@ -221,6 +249,7 @@ class Attachment(dict):
 
     @property
     def name(self) -> str:
+        """name of the attachment (used in filename)"""
         return self["name"]
 
     @classmethod
@@ -270,7 +299,14 @@ class Client(SwaggerClient):
 
     _shared_state = {}
 
-    def __init__(self, apikey=None, headers=None, host=None):
+    def __init__(self, apikey: str = None, headers: dict = None, host: str = None):
+        """Initialize the client - only reloads API spec from server as needed
+
+        Args:
+            apikey (str): API key - can also be set via MPCONTRIBS_API_KEY env var
+            headers (dict): custom headers for localhost connections - ignored if API key set
+            host (str): host address to connect to - can also be set via MPCONTRIBS_API_HOST
+        """
         # - Kong forwards consumer headers when api-key used for auth
         # - forward consumer headers when connecting through localhost
         self.__dict__ = self._shared_state
@@ -298,9 +334,9 @@ class Client(SwaggerClient):
         if "swagger_spec" not in self.__dict__ or (
             self.swagger_spec.http_client.headers != self.headers
         ):
-            self.load()
+            self._load()
 
-    def load(self):
+    def _load(self):
         http_client = FidoClientGlobalHeaders(headers=self.headers)
         loader = Loader(http_client)
         origin_url = f"{self.url}/apispec.json"
@@ -374,26 +410,46 @@ class Client(SwaggerClient):
             swagger_spec, also_return_response=bravado_config.also_return_response
         )
 
-    def get_project(self, project):
-        """Convenience function to get full project entry and display as HTML table"""
-        return Dict(self.projects.get_entry(pk=project, _fields=["_all"]).result())
+    def __dir__(self):
+        members = set(self.swagger_spec.resources.keys())
+        members |= set(k for k in self.__dict__.keys() if not k.startswith("_"))
+        members |= set(k for k in dir(self.__class__) if not k.startswith("_"))
+        return members
 
-    def get_contribution(self, cid):
-        """Convenience function to get full contribution entry"""
-        fields = list(
-            self.swagger_spec.definitions.get("ContributionsSchema")._properties.keys()
-        )  # don't return dynamic fields (card_*)
-        fields.remove("notebook")
+    def _is_valid_payload(self, model: str, data: dict):
+        model_spec = deepcopy(self.get_model(f"{model}sSchema")._model_spec)
+        model_spec.pop("required")
+        model_spec['additionalProperties'] = False
+
+        try:
+            validate_object(self.swagger_spec, model_spec, data)
+        except ValidationError as ex:
+            return str(ex)
+
+        return True
+
+    def get_project(self, name: str) -> Type[Dict]:
+        """Retrieve full project entry
+
+        Args:
+            name (str): name of the project
+        """
+        return Dict(self.projects.get_entry(pk=name, _fields=["_all"]).result())
+
+    def get_contribution(self, cid: str) -> Type[Dict]:
+        """Retrieve full contribution entry
+
+        Args:
+            cid (str): contribution ObjectID
+        """
+        fields = list(self.get_model("ContributionsSchema")._properties.keys())
         return Dict(self.contributions.get_entry(pk=cid, _fields=fields).result())
 
-    def get_table(self, tid_or_md5):
-        """Convenience function to get full Pandas DataFrame for a table.
+    def get_table(self, tid_or_md5: str) -> Type[Table]:
+        """Retrieve full Pandas DataFrame for a table
 
         Args:
             tid_or_md5 (str): ObjectId or MD5 hash digest for table
-
-        Returns:
-            pd.DataFrame: pandas DataFrame containing table data
         """
         str_len = len(tid_or_md5)
         if str_len not in {24, 32}:
@@ -440,14 +496,11 @@ class Client(SwaggerClient):
         ret.attrs = table["attrs"]
         return ret
 
-    def get_structure(self, sid_or_md5):
-        """Convenience function to get pymatgen structure
+    def get_structure(self, sid_or_md5: str) -> Type[Structure]:
+        """Retrieve pymatgen structure
 
         Args:
             sid_or_md5 (str): ObjectId or MD5 hash digest for structure
-
-        Returns:
-            pymatgen.Structure: pymatgen Structure
         """
         str_len = len(sid_or_md5)
         if str_len not in {24, 32}:
@@ -461,9 +514,7 @@ class Client(SwaggerClient):
         else:
             sid = sid_or_md5
 
-        fields = list(
-            self.swagger_spec.definitions.get("StructuresSchema")._properties.keys()
-        )  # don't return dynamic fields (cif)
+        fields = list(self.get_model("StructuresSchema")._properties.keys())
         resp = self.structures.get_entry(pk=sid, _fields=fields).result()
         ret = Structure.from_dict(resp)
         ret.attrs = {
@@ -472,14 +523,11 @@ class Client(SwaggerClient):
         }
         return ret
 
-    def get_attachment(self, aid_or_md5):
-        """Convenience function to get attachment.
+    def get_attachment(self, aid_or_md5: str) -> Type[Attachment]:
+        """Retrieve an attachment
 
         Args:
             aid_or_md5 (str): ObjectId or MD5 hash digest for attachment
-
-        Returns:
-            pd.DataFrame: pandas DataFrame containing table data
         """
         str_len = len(aid_or_md5)
         if str_len not in {24, 32}:
@@ -497,7 +545,7 @@ class Client(SwaggerClient):
 
         return Attachment(self.attachments.get_entry(pk=aid, _fields=["_all"]).result())
 
-    def init_columns(self, project, columns):
+    def init_columns(self, name: str, columns: dict) -> dict:
         """initialize columns for a project to set their order and desired units
 
         The `columns` field tracks the minima and maxima of each `data` field as
@@ -526,11 +574,11 @@ class Client(SwaggerClient):
 
 
         Args:
-            project (str): name of the project for which to initialize data columns
-            columns (dict): dictionary mapping data column to its unit (use None as value)
+            name (str): name of the project for which to initialize data columns
+            columns (dict): dictionary mapping data column to its unit
         """
-        if not isinstance(project, str):
-            return {"error": "`project` argument must be a string!"}
+        if not isinstance(name, str):
+            return {"error": "`name` argument must be a string!"}
 
         if not isinstance(columns, dict):
             return {"error": "`columns` argument must be a dict!"}
@@ -557,7 +605,7 @@ class Client(SwaggerClient):
             if not is_valid_string and v is not None:
                 return {"error": f"Unit '{v}' for {k} invalid (use `None` or a non-NaN string)!"}
 
-            if v != "" and v not in ureg:
+            if v != "" and v is not None and v not in ureg:
                 return {"error": f"Unit '{v}' for {k} invalid!"}
 
             existing_columns.add(k)
@@ -565,14 +613,14 @@ class Client(SwaggerClient):
         resp = self.projects.get_entries(_fields=["name"]).result()
         valid_projects = {p["name"] for p in resp["data"]}
 
-        if project not in valid_projects:
-            return {"error": f"{project} doesn't exist or you don't have access!"}
+        if name not in valid_projects:
+            return {"error": f"{name} doesn't exist or you don't have access!"}
 
         # sort to avoid "overlapping columns" error in handsontable's NestedHeaders
         sorted_columns = flatten(unflatten(columns, splitter="dot"), reducer="dot")
 
         # reconcile with existing columns
-        resp = self.projects.get_entry(pk=project, _fields=["columns"]).result()
+        resp = self.projects.get_entry(pk=name, _fields=["columns"]).result()
         existing_columns, new_columns = {}, []
 
         for col in resp["columns"]:
@@ -615,19 +663,24 @@ class Client(SwaggerClient):
 
             new_columns.append(new_column)
 
-        self.projects.update_entry(pk=project, project={"columns": []}).result()
-        return self.projects.update_entry(
-            pk=project, project={"columns": new_columns}
-        ).result()
+        payload = {"columns": new_columns}
+        valid = self._is_valid_payload("Project", payload)
+        if not valid:
+            return {"error": valid}
 
-    def delete_contributions(self, project, per_page=100, max_workers=5, retry=False):
+        self.projects.update_entry(pk=name, project={"columns": []}).result()
+        return self.projects.update_entry(pk=name, project=payload).result()
+
+    def delete_contributions(
+        self, name: str, per_page: int = 100, max_workers: int = 5, retry: bool = False
+    ):
         """Remove all contributions for a project
 
         Note: This also resets the columns field for a project. It might have to be
-        re-initialized via `client.init_columns()` again.`
+        re-initialized via `client.init_columns()`.
 
         Args:
-            project (str): name of the project for which to delete contributions
+            name (str): name of the project for which to delete contributions
             per_page (int): number of contributions to delete per request
             max_workers (int): maximum number of parallel requests to send at a time
             retry (bool): if True, retry deletion of failed contributions until done
@@ -638,10 +691,10 @@ class Client(SwaggerClient):
             max_workers = MAX_WORKERS
             print(f"max_workers reset to max {MAX_WORKERS}")
 
-        cids = self.get_contributions(project)["ids"]
+        cids = self.get_contributions(name)["ids"]
         total = len(cids)
         # reset columns to be save (sometimes not all are reset BUGFIX?)
-        self.projects.update_entry(pk=project, project={"columns": []}).result()
+        self.projects.update_entry(pk=name, project={"columns": []}).result()
 
         if cids:
             with FuturesSession(max_workers=max_workers) as session:
@@ -651,7 +704,7 @@ class Client(SwaggerClient):
                             f"{self.url}/contributions/",
                             headers=self.headers,
                             params={
-                                "project": project,
+                                "project": name,
                                 "id__in": ",".join(chunk),
                                 "per_page": per_page,
                             },
@@ -660,12 +713,12 @@ class Client(SwaggerClient):
                     ]
 
                     self._run_futures(futures, total=len(cids))
-                    cids = self.get_contributions(project)["ids"]
+                    cids = self.get_contributions(name)["ids"]
 
                     if not retry:
                         break
 
-                self.load()
+                self._load()
 
             toc = time.perf_counter()
             dt = (toc - tic) / 60
@@ -675,28 +728,43 @@ class Client(SwaggerClient):
             else:
                 print(f"It took {dt:.1f}min to delete {total} contributions.")
         else:
-            print(f"There aren't any contributions to delete for {project}")
+            print(f"There aren't any contributions to delete for {name}")
 
     @sleep_and_retry
     @limits(calls=175, period=60)
-    def get_unique_identifiers_flag(self, name):
+    def get_unique_identifiers_flag(self, name: str) -> bool:
+        """Retrieve value of `unique_identifiers` flag for a project
+
+        Args:
+            name (str): name of the project
+        """
         return self.projects.get_entry(
             pk=name, _fields=["unique_identifiers"]
         ).result()["unique_identifiers"]
 
     @sleep_and_retry
     @limits(calls=175, period=60)
-    def get_total_pages(self, name, per_page):
+    def get_total_pages(self, name: str, per_page: int) -> int:
+        """Retrieve total number of pages for contributions in a project
+
+        Args:
+            name (str): name of the project
+            per_page (int): number of contributions per page
+        """
         return self.contributions.get_entries(
             project=name, per_page=per_page, _fields=["id"],
         ).result()["total_pages"]
 
-    def get_contributions(self, name):
-        """get list of existing contributions"""
+    def get_contributions(self, name: str) -> dict:
+        """Retrieve a list of existing contributions and their components for a project
+
+        Args:
+            name (str): name of the project
+        """
         ret = defaultdict(set)
         ret["unique_identifiers"] = self.get_unique_identifiers_flag(name)
         pages = self.get_total_pages(name, 250)
-        id_fields = {"id", "identifier"}
+        id_fields = ["id", "identifier"]
 
         @sleep_and_retry
         @limits(calls=175, period=60)
@@ -708,7 +776,7 @@ class Client(SwaggerClient):
                     "project": name,
                     "page": page,
                     "per_page": 250,
-                    "_fields": ",".join(id_fields | COMPONENTS)
+                    "_fields": ",".join(id_fields + COMPONENTS)
                 },
             )
             setattr(future, "track_id", page)
@@ -738,7 +806,38 @@ class Client(SwaggerClient):
 
         return ret
 
-    def get_number_contributions(self, **query):
+    def update_contributions(self, name: str, data: dict, query: dict = None) -> dict:
+        """Apply the same update to all contributions in a project (matching query)
+
+        See `client.contributions.get_entries()` for keyword arguments used in query.
+
+        Args:
+            name (str): name of the project
+            data (dict): update to apply on every matching contribution
+            query (dict): optional query to select contributions
+        """
+        if not data:
+            return "Nothing to update."
+
+        valid = self._is_valid_payload("Contribution", data)
+        if not valid:
+            return {"error": valid}
+
+        query = query or {}
+        query["project"] = name
+        has_more = True
+        updated = 0
+
+        while has_more:
+            resp = self.contributions.update_entries(
+                contributions=data, _limit=250, **query
+            ).result()
+            has_more = resp["has_more"]
+            updated += resp["count"]
+
+        print(f"Updated {updated} contributions.")
+
+    def get_number_contributions(self, **query) -> int:
         """Retrieve total number of contributions for query
 
         See `client.contributions.get_entries()` for keyword arguments used in query.
@@ -747,16 +846,39 @@ class Client(SwaggerClient):
             _fields=["id"], _limit=1, **query
         ).result()["total_count"]
 
+    def publish(self, name: str, recursive: bool = False) -> dict:
+        """Publish a project and optionally its contributions
+
+        Args:
+            name (str): name of the project
+            recursive (bool): also publish according contributions?
+        """
+        try:
+            resp = self.projects.get_entry(pk=name, _fields=["is_public"]).result()
+        except HTTPNotFound:
+            return {"error": f"project `{name}` not found or access denied!"}
+
+        is_public = resp["is_public"]
+
+        if not recursive and is_public:
+            return {"warning": f"project `{name}` already public (recursive=False)."}
+
+        if not is_public:
+            self.projects.update_entry(pk=name, project={"is_public": True}).result()
+
+        if recursive:
+            self.update_contributions(name, {"is_public": True}, {"is_public": False})
+
     def submit_contributions(
         self,
-        contributions,
-        skip_dupe_check=False,
-        ignore_dupes=False,
-        retry=False,
-        per_page=100,
-        max_workers=3,
+        contributions: list,
+        skip_dupe_check: bool = False,
+        ignore_dupes: bool = False,
+        retry: bool = False,
+        per_page: int = 100,
+        max_workers: int = 3,
     ):
-        """Convenience function to submit a list of contributions
+        """Submit a list of contributions
 
         Example for a single contribution dictionary:
 
@@ -770,7 +892,7 @@ class Client(SwaggerClient):
             },
             "structures": [<pymatgen Structure>, ...],
             "tables": [<pandas DataFrame>, ...],
-            "attachments": [<pathlib.Path>, ...]
+            "attachments": [<pathlib.Path>, <mpcontribs.client.Attachment>, ...]
         }
 
         This function can also be used to update contributions by including the respective
@@ -799,17 +921,26 @@ class Client(SwaggerClient):
         # get existing contributions
         existing = defaultdict(set)
         existing["unique_identifiers"] = True
-
         project_names = set()
         collect_ids = []
+        require_one_of = {"data"} | set(COMPONENTS)
+
         for idx, c in enumerate(contributions):
-            if "id" in c:
+            has_keys = require_one_of & c.keys()
+            if not has_keys:
+                return {"error": f"Nothing to submit for contribution #{idx}!"}
+            elif not all(c[k] for k in has_keys):
+                for k in has_keys:
+                    if not c[k]:
+                        return {"error": f"Empty `{k}` for contribution #{idx}!"}
+            elif "id" in c:
                 collect_ids.append(c["id"])
-            elif "project" in c:
+            elif "project" in c and "identifier" in c:
                 project_names.add(c["project"])
             else:
-                print(f"Fields `project` or `id` missing for contribution #{idx}!")
-                return
+                return {
+                    "error": f"Provide `project` & `identifier`, or `id` for contribution #{idx}!"
+                }
 
         id2project = {}
         if collect_ids:
@@ -834,9 +965,7 @@ class Client(SwaggerClient):
         digests = {project_name: defaultdict(set) for project_name in project_names}
         fields = [
             comp
-            for comp in self.swagger_spec.definitions.get(
-                "ContributionsSchema"
-            )._properties.keys()
+            for comp in self.get_model("ContributionsSchema")._properties.keys()
             if comp not in COMPONENTS
         ]
 
@@ -960,6 +1089,10 @@ class Client(SwaggerClient):
                     digests[project_name][component].add(digest)
                     contribs[project_name][-1][component].append(dct)
 
+                valid = self._is_valid_payload("Contribution", contribs[project_name][-1])
+                if not valid:
+                    return {"error": f"{contrib['identifier']} invalid: {valid}!"}
+
         # submit contributions
         if contribs:
             print("submit contributions ...")
@@ -1023,11 +1156,74 @@ class Client(SwaggerClient):
                 )
                 toc = time.perf_counter()
                 dt = (toc - tic) / 60
-                self.load()
+                self._load()
                 print(f"It took {dt:.1f}min to submit {updated_total} contributions.")
         else:
             print("Nothing to submit.")
 
+    def download_contribution(
+        self,
+        cid: str,
+        outdir: Union[str, Path] = "mpcontribs-downloads",
+        overwrite: bool = False,
+        include: list = []
+    ) -> Path:
+        """Download a single contribution as .json.gz file(s)
+
+        Args:
+            cid: contribution ObjectId
+            outdir: optional existing output directory
+            overwrite: force re-download of existing contribution/components
+            include: components to include in download
+        """
+        outdir = Path(outdir) or Path(".")
+        outdir.mkdir(parents=True, exist_ok=True)
+        if any(x for x in include if x not in COMPONENTS):
+            print(f"`include` must be subset of {COMPONENTS}!")
+            return
+
+        include = set(x for x in include if x in COMPONENTS)
+        fields = ["project"] + list(include)
+        contrib = self.contributions.get_entry(pk=cid, _fields=fields).result()
+        components = []
+
+        if include:
+            proj = self.projects.get_entry(pk=contrib["project"], _fields=["columns"]).result()
+            for column in proj["columns"]:
+                path = column["path"]
+                if path.split(".", 1)[0] in include:
+                    components.append(path)
+
+        model = self.get_model("ContributionsSchema")
+        fields = list(k for k in model._properties.keys() if k not in COMPONENTS) + components
+
+        # download contribution
+        subdir = outdir / cid
+        subdir.mkdir(exist_ok=True)
+        path = subdir / "contribution.json.gz"
+
+        if not path.exists() or overwrite:
+            content = self.contributions.download_entries(
+                id=cid, short_mime="gz", format="json", _fields=fields
+            ).result()
+            path.write_bytes(content)
+            print(path)
+
+        # download components
+        for component in components:
+            path = subdir / f"{component}.json.gz"
+            if not path.exists() or overwrite:
+                ids = [x["id"] for x in contrib[component]]
+                model = self.get_model(f"{component.capitalize()}Schema")
+                fields = list(model._properties.keys())
+                content = getattr(self, component).download_entries(
+                    id__in=ids, short_mime="gz", format="json", _fields=fields
+                ).result()
+                path.write_bytes(content)
+                print(path)
+
+        return subdir
+
     def _run_futures(self, futures, total=None):
         """helper to run futures/requests"""
         responses = {}
