@@ -31,6 +31,7 @@ def test_Client():
     assert spec.spec_dict["host"] == DEFAULT_HOST
     assert spec.spec_dict["schemes"] == ["https"]
     assert spec.user_defined_formats["email"] == email_format
+
     kwargs = {"headers": {"a": "b"}, "host": "localhost:5000"}
     spec = Client(**kwargs).swagger_spec
     assert spec.http_client.headers == {"a": "b"}
@@ -39,6 +40,26 @@ def test_Client():
     assert spec.spec_dict["schemes"] == ["http"]
     assert spec.user_defined_formats["email"] == email_format
 
+    kwargs = {"host": "contribs-api:5000"}
+    spec = Client(**kwargs).swagger_spec
+    assert spec.http_client.headers == {}
+    assert spec.origin_url == "http://contribs-api:5000/apispec.json"
+    assert spec.spec_dict["host"] == "contribs-api:5000"
+    assert spec.spec_dict["schemes"] == ["http"]
+    assert spec.user_defined_formats["email"] == email_format
+
+    kwargs = {"host": "ml-api.materialsproject.org"}
+    spec = Client(**kwargs).swagger_spec
+    assert spec.http_client.headers == {}
+    assert spec.origin_url == "https://ml-api.materialsproject.org/apispec.json"
+    assert spec.spec_dict["host"] == "ml-api.materialsproject.org"
+    assert spec.spec_dict["schemes"] == ["https"]
+    assert spec.user_defined_formats["email"] == email_format
+
+    with pytest.raises(ValueError):
+        kwargs = {"host": "not.valid.org"}
+        spec = Client(**kwargs).swagger_spec
+
 
 def test_Client_Live():
     Client()
